/* Terraform - (C) 1997-1999 Robert Gasch (r.gasch@chello.nl)
 *  - http://212.187.12.197/RNG/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "HeightField.h" 
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"



int	HeightField::s_HFcount = 0;


/*
 *  constructor: take a hf and build the object from it
 */
HeightField::HeightField (PTYPE *hf, int xsize, int ysize, char *name) 
	    : HeightFieldCore (hf, xsize, ysize)
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ HeightField\n");
	p_name = NULL;
	init (this->p_hf, xsize, ysize, name);
}


/*
 *  constructor: create a new HF of specified size 
 */
HeightField::HeightField (int xsize, int ysize, char *name) 
		: HeightFieldCore ((xsize*ysize==0?NULL:new PTYPE[xsize*ysize]), xsize, ysize)
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ HeightField\n");
	init (this->p_hf, xsize, ysize, name);
	
}


/*
 *  init: initialize the object properties according to the parameters 
 */
void HeightField::init (PTYPE *hf, int xsize, int ysize, char *name)
{
	// Sanity check if settings are OK 
	if (hf)
		{
		d_updateTime = time(NULL);
		SanityCheck::bailout ((xsize<1), "xsize<1", "HeightField::HeightField");
		SanityCheck::bailout ((ysize<1), "ysize<1", "HeightField::HeightField");
		}
	else
		{
		SanityCheck::bailout ((xsize), "xsize!=0", "HeightField::HeightField");
		SanityCheck::bailout ((ysize), "ysize!=0", "HeightField::HeightField");
		d_updateTime = 0;
		}

	// get the right name
	if ((p_name && name) || name)
		{
		if (p_name != name)
			{
			if (p_name)
				delete p_name;
			p_name = strdup (name);
			}
		}
	else
		p_name = getNewName (); 

	// initialized the data members
	p_hfBck=NULL;
	p_contourList=NULL;
	b_loadedFromFile=FALSE;
	b_isSaved=TRUE;
	d_backupTime = 0;
}



/*
 *  destructor: clean up, free allocated memory
 */
HeightField::~HeightField ()
{
	if (p_name)
		free (p_name);
	destroyBackup ();

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- HeightField\n");
}


/*
 *  setName: assign a new (non-NULL) Name 
 */
void HeightField::setName (char *name)
{
	SanityCheck::bailout ((!name), "name==NULL", "HeightField::setName");

	if (p_name)
		free (p_name);
	p_name = strdup (name);
}



/*
 *  setSaved: set the b_isSaved flag
 */
void HeightField::setSaved (bool isSaved)
{
	b_isSaved = isSaved;
	d_updateTime = time(NULL);

	if (!isSaved && p_contourList)		// expire contour list
		setContourList (NULL, time (NULL));

	gatherStatistics ();
}


/*
 *  setContourList: assign a new contour list and delete the old one
 */
void HeightField::setContourList(TFGListFlexArr *l, time_t timestamp)
{
	// delete old list
	if (p_contourList)
		delete p_contourList;

	p_contourList = l;
	d_clTimestamp = timestamp;
}


/*
 *  getNewName: return a newly allocated string for the next name.
 * 	receiver becomes the owner of this piece of memory!
 */
char *HeightField::getNewName ()
{
	char	buf[256];
	char	*newName;

	sprintf (buf, "%s-%02d.tga", TF_DEFAULT_FILENAME, ++s_HFcount);
	newName = strdup (buf);
	return (newName);
}


/*
 *  getCopyOfData: return a newly allocated copy of the actual data stream. 
 * 	The receiver becomes the owner of this piece of memory!
 */
PTYPE *HeightField::getCopyOfData ()
{
	SanityCheck::bailout ((!p_hf), "p_hf==NULL", "HeightField::getCopyOfData");

	PTYPE	 *hfnew = new PTYPE[d_size];

	memcpy (hfnew, this->p_hf, this->d_size*sizeof(PTYPE));
	return hfnew;
}


/*
 *  releaseData: release the data pointer and reset everything ... 
 * 	If you don't keep track of this pointer, you're going to have 
 * 	some really large memory leaks!
 */
PTYPE *HeightField::releaseData ()
{
	SanityCheck::bailout ((!p_hf), "p_hf==NULL", "HeightField::releaseData");
	PTYPE	*data = p_hf;

	init (NULL, 0, 0, "empty");
	return data;
}


/*
 *  newHF: discard the current HF and allocate a new one of specified size 
 */
int HeightField::newHF (int xsize, int ysize)
{
	int	size = xsize * ysize;
	PTYPE 	*hf = new PTYPE[size];

	
	memset (hf, size, 0);
	b_isSaved = FALSE;
	this->setData (hf, xsize, ysize);
	destroyBackup ();
	return 0;
}


/*
 *  backup: make a backup copy of the current height field
 */
void HeightField::backup ()
{
	SanityCheck::bailout ((!this->p_hf), "this->p_hf==NULL", "HeightField::backup");

	if (p_hfBck)
		delete [] p_hfBck;
	p_hfBck = new PTYPE[d_size];
	memcpy (p_hfBck, this->p_hf, sizeof(PTYPE)*this->d_size);
	p_hfBckSealevel = d_sealevel;
	d_backupTime = d_updateTime;
}


/*
 *  restore: restore a backup copy of the current height field. Doesn't
 *  	     calculate statistics! 
 */
void HeightField::restoreBackup ()
{
	SanityCheck::bailout ((!this->p_hf), "this->p_hf==NULL", "HeightField::restore");
	SanityCheck::bailout ((!p_hfBck), "p_hfBck==NULL", "HeightField::restore");

	memcpy (this->p_hf, p_hfBck, sizeof(PTYPE)*this->d_size);
	d_sealevel = p_hfBckSealevel;
	d_updateTime = d_backupTime;
}


/*
 *  destroyBackup: destroy the backup copy of the current height field and 
 * 	free the memory. 
 */
void HeightField::destroyBackup ()
{
	if (!p_hfBck)
		return;

	delete [] p_hfBck;
	p_hfBck = NULL;
	d_backupTime = 0;
}


